﻿
using UnityEngine;
using UnityEngine.Rendering;

namespace PluginIdea.VolumeLight
{
    public class VolumeLightGeometry : MonoBehaviour
    {
        private VolumetricLightMono volumetricLight;
        private Material geometryMaterial = null;

        public MeshRenderer Renderer { get; private set; }
        public MeshFilter Filter { get; private set; }
        public Mesh ConeMesh { get; private set; }

        public float geoRadius = 1;

        private void Start()
        {
            if (!volumetricLight)
                DestroyImmediate(gameObject);
        }

        private void OnDestroy()
        {
            if (geometryMaterial)
            {
                DestroyImmediate(geometryMaterial);
                geometryMaterial = null;
            }
        }

        public void AppendEssensialComponents(VolumetricLightMono lightMono)
        {
            volumetricLight = lightMono;

            transform.SetParent(lightMono.transform, false);
            Renderer = Utils.GetSafeComponent<MeshRenderer>(gameObject);

            geometryMaterial = CreateMaterial();
            ApplyMaterialProperties();

            Filter = Utils.GetSafeComponent<MeshFilter>(gameObject);
        }

        public void RegenerateMesh(int sides)
        {
            gameObject.layer =1;

            ConeMesh = Utils.GetConeMesh(sides);
            Filter.sharedMesh = ConeMesh;

            SetMaterialPropertys();
        }

        private void SetUpLocalScale()
        {
            float maxRadius = Mathf.Max(volumetricLight.startConeRadius, volumetricLight.endConeRadius);
            transform.localScale = new Vector3(maxRadius, maxRadius, volumetricLight.lightFadeEnd);
        }

        private void SetMaterialProperty(string propertyName, float value) => SetMaterialPropertyInternal(propertyName, value);
        private void SetMaterialProperty(string propertyName, Vector4 value) => SetMaterialPropertyInternal(propertyName, value);
        private void SetMaterialProperty(string propertyName, Color value) => SetMaterialPropertyInternal(propertyName, value);

        private void SetMaterialPropertyInternal(string propertyName, object value)
        {
            if (geometryMaterial)
            {
                if (value is float)
                {
                    geometryMaterial.SetFloat(propertyName, (float)value);
                }
                else if (value is Vector4)
                {
                    geometryMaterial.SetVector(propertyName, (Vector4)value);
                }
                else if (value is Color)
                {
                    geometryMaterial.SetColor(propertyName, (Color)value);
                }
            }
        }

        private bool ApplyMaterialProperties()
        {
            Renderer.material = geometryMaterial;
            return geometryMaterial != null;
        }

        public void SetMaterialPropertys()
        {
            if (!ApplyMaterialProperties()) return;

            // Prepare some calculation parameters
            SetMaterialProperty("_A", volumetricLight.spotlightAngle * 0.0086f);
            SetMaterialProperty("_RadiusBeginEnd", new Vector2(volumetricLight.startConeRadius, volumetricLight.endConeRadius));

            // When there are multiple beams at the same time, the material parameters can be set separately without affecting each other.
            SetMaterialProperty("_MainColor", volumetricLight.mainColor);
            SetMaterialProperty("_IntensityMul", volumetricLight.intensityFactor * volumetricLight.lightIntensity);
            SetMaterialProperty("_AlphaIN", volumetricLight.internalAlpha);
            SetMaterialProperty("_AlphaOut", volumetricLight.externalAlpha);
            SetMaterialProperty("_FadeStartDis", volumetricLight.lightFadeStart);
            SetMaterialProperty("_FadeEndDis", volumetricLight.lightFadeEnd);

            Shader.SetGlobalTexture("noise3DTex", HPVLB.dic_name2Object["NoiseTex"]);

            SetMaterialProperty("_NoiseSpeed", new Vector4(volumetricLight.noiseSpeed.x, volumetricLight.noiseSpeed.y, volumetricLight.noiseSpeed.z, volumetricLight.noiseScale));
            SetMaterialProperty("_NoiseData", new Vector3(volumetricLight.noiseStrength, 0, 0));

            if (geometryMaterial == null)
                Renderer.SetPropertyBlock(materialPropertyBlock);

            SetUpLocalScale();
        }

        private void OnWillRenderObject()
        { 
            // Natural transition at the junction ensures that it works in every camera.
            var camera = Camera.current;
            if (camera != null)
            {
                if (camera && volumetricLight && geometryMaterial)
                {
                    Renderer.GetPropertyBlock(materialPropertyBlock);
                    Renderer.SetPropertyBlock(materialPropertyBlock);

                    camera.depthTextureMode |= DepthTextureMode.Depth;
                }
            }
        }

        private Material CreateMaterial()
        {
            var shader = HPVLB.dic_name2shader["PluginIdea/HPVL"];
            var material = new Material(shader);
            material.renderQueue = 3100;
            material.enableInstancing = true;
            return material;
        }

        private MaterialPropertyBlock materialPropertyBlock = new MaterialPropertyBlock();
    }
}